from starlette.responses import JSONResponse
from starlette.endpoints import WebSocketEndpoint, HTTPEndpoint
import websockets
from ..settings import gremlin_server_url, gremlin_traversal_source, gremlin_server_username, gremlin_server_password
import json
import uuid
import logging
import base64


class GremlinWebsocketHandler(object):
    conn = None

    async def connect(self, url):
        self.conn = await websockets.connect(url)

    async def send(self, msg):
        await self.conn.send(msg)

    async def recv(self):
        return await self.conn.recv()

    async def close(self):
        await self.conn.close()


class HomePageView(HTTPEndpoint):

    async def get(self, request):
        return JSONResponse({'message': 'Hello world! go to /graphql'})


class GremlinQueryView(WebSocketEndpoint):
    encoding = 'json'
    gremlin = GremlinWebsocketHandler()

    async def on_connect(self, websocket):
        logging.debug("connected")
        await websocket.accept()
        await self.gremlin.connect(gremlin_server_url)

    @staticmethod
    async def prepare_message(gremlin_query):
        request_id = str(uuid.uuid4())
        query_message = {
            "requestId": request_id,
            "args": {
                "gremlin": gremlin_query,
                "bindings": {},
                "language": "gremlin-groovy",
                "aliases": {"g": gremlin_traversal_source},
                "session": request_id
            }
        }
        if gremlin_server_username or gremlin_server_password:
            auth = b"".join([b"\x00", gremlin_server_username.encode("utf-8"),
                             b"\x00", gremlin_server_password.encode("utf-8")])
            query_message['op'] = "authentication"
            query_message['processor'] = ""
            query_message['args']['sasl'] = base64.b64encode(auth).decode()
        else:
            query_message['op'] = "eval"
            query_message['processor'] = "session"
            query_message['args']['session'] = request_id
        return query_message

    async def on_receive(self, websocket, data):
        gremlin_query = data.get("gremlin")
        query_message = await self.prepare_message(gremlin_query)
        query_string = json.dumps(query_message)
        await self.gremlin.send(query_string)
        while True:
            try:
                response_data = await self.gremlin.recv()
                print(response_data)
                await websocket.send_json(json.loads(response_data))
            except websockets.exceptions.ConnectionClosed:
                logging.debug('Don\'t have any more messages. closed connection with gremlin server')
                # await self.gremlin.close()
                break

    async def on_disconnect(self, websocket, close_code):
        logging.debug("disconnected")
        await self.gremlin.close()



import logging

from gremlin_python.structure.graph import Vertex, Edge, Path


class GraphSONV3Reader:

    @staticmethod
    def get_element_id(_id):
        if type(_id) is dict:
            _id = _id["@value"]
        if type(_id) is dict:
            _id = _id['relationId']
        if type(_id) is str:
            _id = _id.strip("#")
        return _id

    @staticmethod
    def check_element_type(_id):
        if type(_id) is dict:
            _id = _id["@value"]
        if type(_id) is dict:
            _id = _id['relationId']
            return "edge"
        return "vertex"

    def serialize_element_dict(self, elem):
        cleaned_data = {"properties": {}}
        # if "Direction.IN" in elem.keys():
        #     cleaned_data['type'] = "g:Edge"
        # else:
        #     cleaned_data['type'] = "g:Vertex"

        for k, v in elem.items():
            if str(k) == "T.id":
                if self.get_element_id(v):
                    cleaned_data['id'] = self.get_element_id(v)

                    if self.check_element_type(v) == "edge":
                        cleaned_data['type'] = "g:Edge"
                    else:
                        cleaned_data['type'] = "g:Vertex"
                else:
                    raise Exception("This element doesnt have id, ( may be path)")

            elif str(k) == "T.label":
                cleaned_data['label'] = v
            elif str(k) == "Direction.OUT":
                _ = self.serialize_element_dict(v)
                cleaned_data.update({"outV": _['id'], "outVLabel": _['label']})
            elif str(k) == "Direction.IN":
                _ = self.serialize_element_dict(v)
                cleaned_data.update({"inV": _['id'], "inVLabel": _['label']})
            else:
                # TODO - check if this is right.
                if type(v) is list:
                    if v.__len__() > 0:
                        if not isinstance(v[0], Edge) or not isinstance(v[0], Vertex):
                            cleaned_data['properties'][k] = v[0]
                        else:
                            raise Exception("This element is not a element dictionary, ( may be path)")
                elif isinstance(v, Edge):
                    raise Exception("This element is not a element dictionary, Edge found in iter, ( may be path)")
                elif isinstance(v, Vertex):
                    raise Exception("This element is not a element dictionary Vertex found in iter, ( may be path)")
                else:
                    cleaned_data['properties'][k] = v

        # if cleaned_data['properties'].keys().__len__() == 0:
        #     del cleaned_data['properties']
        return cleaned_data

    @staticmethod
    def serialize_vertex_element(vertex):
        return {
            "id": vertex.id,
            "label": vertex.label,
            "type": "g:Vertex",
            "properties": {}
        }

    def serialize_edge_element(self, edge):

        _ = {
            "id": self.get_element_id(edge.id),
            "label": edge.label,
            "type": "g:Edge",
            "in_v": edge.inV.id,
            "in_v_label": edge.inV.label,
            "out_v": edge.outV.id,
            "out_v_label": edge.outV.label,
            "properties": {}
        }
        return _

    def serialize_data(self, data):
        if isinstance(data, list):
            _serialized_data = []
            for datum in data:
                _ = self.serialize_data(datum)
                if type(_) is list:
                    _serialized_data.extend(_)
                else:
                    _serialized_data.append(_)
            return _serialized_data

        elif isinstance(data, dict):
            keys = list(data.keys())
            if "label" in keys and "type" in keys:
                # TODO - find a better way to distinguish already invana serialised element.
                return data
            else:
                try:
                    _ = self.serialize_element_dict(data)
                    return _
                except Exception as e:
                    logging.debug(e)
                    _serialised_data = []
                    for key, val in data.items():
                        if type(val) is list:
                            _serialised_data.extend(self.serialize_data(val))
                        else:
                            _serialised_data.append(self.serialize_data(val))
                    return _serialised_data
        elif isinstance(data, Vertex):
            return self.serialize_vertex_element(data)
        elif isinstance(data, Edge):
            return self.serialize_edge_element(data)
        elif isinstance(data, Path):
            _serialized_data = []
            for datum in data:
                _ = self.serialize_data(datum)
                _serialized_data.append(_)

            # add edge inv and outv info to the edge
            _serialized_data[1]['out_v'] = _serialized_data[0]['id']
            _serialized_data[1]['out_v_label'] = _serialized_data[0]['label']
            _serialized_data[1]['in_v'] = _serialized_data[2]['id']
            _serialized_data[1]['in_v_label'] = _serialized_data[2]['label']
            return _serialized_data
        else:
            raise NotImplementedError("")



from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.routing import Route, WebSocketRoute
from starlette.graphql import GraphQLApp
from invana_engine.server.schemas.query import GremlinQuery
from .schemas.mutation import GremlinMutation
from graphene import Schema
from .views import HomePageView, GremlinQueryView
from invana_engine.gremlin import InvanaEngineClient
import time
from ..settings import gremlin_server_url, gremlin_server_password, gremlin_server_username, shall_debug, \
    gremlin_traversal_source, invana_engine_version
from termcolor import cprint

print(".......................................................")
cprint(f"Invana Engine Server {invana_engine_version}", "cyan", attrs=["bold"])
print(".......................................................")
print(f"Using GREMLIN_SERVER_URL: {gremlin_server_url}")
print(f"Using GREMLIN_TRAVERSAL_SOURCE: {gremlin_traversal_source}")
if gremlin_server_username:
    print("Using GREMLIN_SERVER_USERNAME: ******** ")
if gremlin_server_password:
    print("Using GREMLIN_SERVER_PASSWORD: ******** ")
print(f"Using DEBUG: {shall_debug}")
print(".......................................................")

if gremlin_server_url is None:
    print("ERROR: GREMLIN_SERVER_URL environment variable not set. Please fix it .")
    print("Exiting the program now. Please refer the documentation at https://github.com/invanalabs/invana-engine")
    exit()

routes = [
    Route('/', HomePageView),
    # Route('/gremlin', endpoint=query_gremlin, methods=['POST']),
    WebSocketRoute('/gremlin', GremlinQueryView),
    Route('/graphql', GraphQLApp(
        schema=Schema(query=GremlinQuery, mutation=GremlinMutation),
    ))
]

middleware = [
    Middleware(CORSMiddleware, allow_origins=['*'], allow_methods=["GET", "POST", "PUT", "DELETE"])
]

app = Starlette(routes=routes, middleware=middleware, debug=shall_debug)
time.sleep(1)
gremlin_client = InvanaEngineClient(
    gremlin_server_url=gremlin_server_url,
    gremlin_server_username=gremlin_server_username,
    gremlin_server_password=gremlin_server_password,
    gremlin_traversal_source=gremlin_traversal_source
)
app.state.gremlin_client = gremlin_client
